/*
** kernel.c                 Kernel access functions to retrieve user number
**
** This program is in the public domain and may be used freely by anyone
** who wants to. 
**
** Last update: 17 Aug 1992
**
** Please send bug fixes/bug reports to: Peter Eriksson <pen@lysator.liu.se>
*/

#ifdef SVR4
#  define STRNET
#endif

#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <nlist.h>
#include <pwd.h>
#include <signal.h>
#include <syslog.h>

#ifdef HAVE_KVM
#  include <kvm.h>
#else
#  include "kvm.h"
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/socket.h>

#ifdef __convex__
#  define _KERN_UTIL
#endif

#include <sys/socketvar.h>

#ifdef __convex__
#  undef _KERN_UTIL
#endif

#if !defined(hpux) && !defined(__hpux)
#  ifdef IRIX
#    define _KERNEL
#  else
#    define KERNEL
#  endif
#endif

#include <sys/file.h>

#ifdef IRIX
#  include <sys/inode.h>
#endif


#include <fcntl.h>

#ifdef ultrix
#  include <sys/dir.h>
#  undef KERNEL
#endif

#ifdef SVR4
#  include <unistd.h>
#  include <sys/fcntl.h>
#  include <sys/stream.h>
#  include <sys/stat.h>
#  include <sys/cred.h>
#  include <sys/vnode.h>
#  include <sys/mkdev.h>
#  define ucred cred
#else
#  include <sys/user.h>
#endif

#include <sys/wait.h>
  
#if !defined(hpux) && !defined(__hpux)
#  ifdef IRIX
#    undef _KERNEL
#  else
#    undef KERNEL
#  endif
#endif

#include <net/if.h>
#include <net/route.h>
#include <netinet/in.h>
#include <netinet/in_pcb.h>
#include <netinet/tcp.h>
#include <netinet/ip_var.h>
#include <netinet/tcp_timer.h>
#include <netinet/tcp_var.h>

#ifndef HPUX7
#  include <arpa/inet.h>
#endif

#ifdef MIPS
#  include <sysv/sys/var.h>
extern int errno;
#endif

#include "identd.h"
#include "error.h"


extern void *calloc();
extern void *malloc();


struct nlist nl[] =
{
#if (defined(hp9000s800) && !defined(HPUX7)) || \
    (defined(ultrix) && defined(mips)) || \
    (defined(sun) && defined(i386))

  /* HP9000 Series 800, HP-UX 8.0,
     Ultrix on DEC's MIPS machines or
     SunOS 4.0.2 on Sun 386i */

#define N_FILE  0  
#define N_NFILE 1
#define N_TCB   2

  { "file" },
  { "nfile" },
  { "tcb" },
  
#else
#if defined(hp9000s800) && defined(HPUX7)
      
  /* HP9000 Series 800, HP-UX 7.0 */
      
#define N_FILE  0  
#define N_NFILE 1
#define N_TCB   2
      
  { "file" },
  { "nfile" },
  { "tcp_cb" },
#else
#ifdef SVR4		/* System V/Release 4 */

#define N_FILE  0
#define N_TCB 1
      
  { "file" },
  { "tcb" },
          
#else
#ifdef HPUX7		/* HP-UX 7.0 */

#define N_FILE  0  
#define N_NFILE 1
#define N_TCB   2

  { "_file" },
  { "_nfile" },
  { "_tcp_cb" },

#else
#ifdef MIPS		/* MIPS RISC/OS */

#define N_FILE 0  
#define N_V    1
#define N_TCB  2

  { "_file" },
  { "_v" },
  { "_tcb" },

#else
#ifdef IRIX	        /* SGI IRIX */

#define N_FILE 0  
#define N_V    1
#define N_TCB  2
 
  { "file" },
  { "v" },
  { "tcb" },
 
#else			/* SUN or BSD or other systems */

#define N_FILE 0  
#define N_NFILE 1
#define N_TCB 2

  { "_file" },
  { "_nfile" },
  { "_tcb" },

#endif
#endif
#endif
#endif
#endif
#endif
  { "" }
};


static kvm_t *kd;

#ifndef SVR4
static struct file *xfile;
static int nfile;
#endif

#if defined(MIPS) || defined(IRIX)
static struct var v;
#endif

static struct inpcb tcb;


int k_open()
{
  /*
  ** Open the kernel memory device
  */
  if (!(kd = kvm_open(path_unix, path_kmem, NULL, O_RDONLY, NULL)))
    ERROR("main: kvm_open");
  
  /*
  ** Extract offsets to the needed variables in the kernel
  */
  if (kvm_nlist(kd, nl) != 0)
    ERROR("main: kvm_nlist");

  return 1;
}


/*
** Get a piece of kernel memory with error handling.
** Returns 1 if call succeeded, else 0 (zero).
*/
static int getbuf(addr, buf, len, what)
  long addr;
  char *buf;
  int len;
  char *what;
{

  if (kvm_read(kd, addr, buf, len) < 0)
  {
    if (syslog_flag)
      syslog(LOG_ERR, "getbuf: kvm_read(%08x, %d) - %s : %m",
	     addr, len, what);

    return 0;
  }
  
  return 1;
}



/*
** Traverse the inpcb list until a match is found.
** Returns NULL if no match.
*/
static
#ifdef SVR4
  short
#else  
  struct socket *
#endif
    getlist(pcbp, faddr, fport, laddr, lport)
  struct inpcb *pcbp;
  struct in_addr *faddr;
  int fport;
  struct in_addr *laddr;
  int lport;
{
  struct inpcb *head;
  

  if (!pcbp)
#ifdef SVR4
    return -1;
#else
    return NULL;
#endif
  
  head = pcbp->inp_prev;
  do 
  {
    if ( pcbp->inp_faddr.s_addr == faddr->s_addr &&
	 pcbp->inp_laddr.s_addr == laddr->s_addr &&
	 pcbp->inp_fport        == fport &&
	 pcbp->inp_lport        == lport )
#ifdef SVR4
      return pcbp->inp_minor;
#else
      return pcbp->inp_socket;
#endif

  } while (pcbp->inp_next != head &&
	   getbuf((long) pcbp->inp_next,
		  pcbp,
		  sizeof(struct inpcb),
		  "tcblist"));

#ifdef SVR4
  return -1;
#else
  return NULL;
#endif
}



/*
** Return the user number for the connection owner
*/
int k_getuid(faddr, fport, laddr, lport, uid)
  struct in_addr faddr;
  int fport;
  struct in_addr laddr;
  int lport;
  int *uid;
{
  long addr;
#ifdef SVR4
  struct stat s;
  short tcpmajor;	/* Major number of the tcp device */
  short sockp;		/* Really a device minor number */
  struct file *filehead, file;
#else
  struct socket *sockp;
  int i;
#endif
#ifdef IRIX
  struct inode inode;
#else
  struct ucred ucb;
#endif
  

  /* -------------------- FILE DESCRIPTOR TABLE -------------------- */
#if defined(MIPS) || defined(IRIX)
  if (!getbuf(nl[N_V].n_value, &v, sizeof(v), "v"))
    return -1;
  
  nfile = v.v_file;
  addr = nl[N_FILE].n_value;
#else
#ifndef SVR4
  if (!getbuf(nl[N_NFILE].n_value, &nfile, sizeof(nfile), "nfile"))
    return -1;
#endif
  if (!getbuf(nl[N_FILE].n_value, &addr, sizeof(addr), "&file"))
    return -1;
#endif

#ifdef SVR4
  /* we just need to determine the major number of the tcp module, which
     is the minor number of the tcp clone device /dev/tcp */
  if (stat("/dev/tcp", &s))
    ERROR("stat(\"/dev/tcp\")");
  
  tcpmajor = minor(s.st_rdev);
#else
  xfile = (struct file *) calloc(nfile, sizeof(struct file));
  if (!xfile)
    ERROR2("k_getuid: calloc(%d,%d)", nfile, sizeof(struct file));
  
  if (!getbuf(addr, xfile, sizeof(struct file)*nfile, "file[]"))
    return -1;
#endif

  /* -------------------- TCP PCB LIST -------------------- */
  if (!getbuf(nl[N_TCB].n_value, &tcb, sizeof(tcb), "tcb"))
    return -1;
  tcb.inp_prev = (struct inpcb *) nl[N_TCB].n_value;
  sockp = getlist(&tcb, faddr, fport, laddr, lport);
#ifdef SVR4
  if (sockp < 0)
#else
  if (!sockp)
#endif
    return -1;

#ifndef SVR4
  /*
  ** Locate the file descriptor that has the socket in question
  ** open so that we can get the 'ucred' information
  */
  for (i = 0; i < nfile; i++)
  {
    if (xfile[i].f_count == 0)
      continue;
    
#ifndef IRIX      
    if (xfile[i].f_type == DTYPE_SOCKET &&
	(struct socket *) xfile[i].f_data == sockp)
    {
      if (!getbuf(xfile[i].f_cred, &ucb, sizeof(ucb)))
	return -1;
      
      *uid = ucb.cr_ruid;
      return 0;
    }
#else /* IRIX */
    if (!getbuf(xfile[i].f_inode, &inode, sizeof(inode)))
      return -1;
    
    if ((inode.i_ftype & IFMT) == IFCHR	&&
	soc_fsptr(&inode) == sockp)
    {
      *uid = inode.i_uid;
      return 0;
    }
#endif
  }
  
#else /* We are SVR4 */

  /*
  * Get the pointer to the head of the file list
  */
  if (!getbuf(nl[N_FILE].n_value, &filehead, sizeof(struct file *)))
    return -1;
  
  file.f_next = filehead;
  do
  {
    struct vnode vnode;
 
    /*
    ** Grab a file record
    */
    if (!getbuf((long)file.f_next, &file, sizeof(struct file), "file"))
      return -1;

    /*
    ** Grab the vnode information to check whether this file is the
    ** correct major & minor char device
    */
    if (!getbuf((off_t)file.f_vnode, &vnode, sizeof(struct vnode), "vnode"))
      return -1;
    
    if ((vnode.v_type == VCHR) && (major(vnode.v_rdev) == tcpmajor) &&
	(minor(vnode.v_rdev) == sockp))
    {
      /*
      ** We've found it!
      */
      if (!getbuf((off_t)file.f_cred, &ucb, sizeof(struct cred), "cred"))
	return -1;
      
      *uid = ucb.cr_ruid;
      return 0;
    }
  } while (file.f_next != filehead);
#endif
  
  return -1;
}

